open OUnit
open Looping
open Block

let parsed s =
  let lexbuf = Lexing.from_string s in
  let (sch,_) = Parser.main Lexer.token lexbuf in
  sch

let sch1 = parsed "/\\i=1..n P_i"
let bck1 = Block.of_schs_and_cstr ~schs:[sch1] ~cstr:Constraint.top
let sch1' = parsed "/\\i=1..n-1 P_i"
let bck1' = Block.of_schs_and_cstr ~schs:[sch1'] ~cstr:Constraint.top
let sch2 = parsed "/\\i=1..n+1 P_i"
let bck2 = Block.of_schs_and_cstr ~schs:[sch2] ~cstr:Constraint.top
let sch2' = sch1
let bck2' = Block.of_schs_and_cstr ~schs:[sch2'] ~cstr:Constraint.top
let sch1'' = parsed "/\\i=1..n-2 P_i"
let bck1'' = Block.of_schs_and_cstr ~schs:[sch1''] ~cstr:Constraint.top

let sch3 = parsed "/\\i=1..n (~P_i \\/ P_i+1)"
let bck3 = Block.of_schs_and_cstr ~schs:[sch3] ~cstr:Constraint.top
let sch3' = parsed "/\\i=1..n-1 (~P_i \\/ P_i+1)"
let bck3' = Block.of_schs_and_cstr ~schs:[sch3'] ~cstr:Constraint.top
let sch4 = parsed "/\\i=1..n+1 (~P_i \\/ P_i+1)"
let bck4 = Block.of_schs_and_cstr ~schs:[sch4] ~cstr:Constraint.top
let sch4' = sch3
let bck4' = Block.of_schs_and_cstr ~schs:[sch4'] ~cstr:Constraint.top
let sch3'' = parsed "/\\i=1..n-2 (~P_i \\/ P_i+1)"
let bck3'' = Block.of_schs_and_cstr ~schs:[sch3''] ~cstr:Constraint.top

let schA = parsed "P_n+1"
let bckA = Block.of_schs_and_cstr ~schs:[sch1;schA] ~cstr:Constraint.top
let schA' = parsed "P_n"
let bckA' = Block.of_schs_and_cstr ~schs:[sch1';schA'] ~cstr:Constraint.top

let schB = parsed "~P_n+1"
let bckB = Block.of_schs_and_cstr ~schs:[sch3;schB] ~cstr:Constraint.top
let schB' = parsed "~P_n"
let bckB' = Block.of_schs_and_cstr ~schs:[sch3';schB'] ~cstr:Constraint.top

let schC = parsed "P_1"
let bckC  = Block.of_schs_and_cstr ~schs:[schB;sch3;schC]   ~cstr:Constraint.top
let bckC' = Block.of_schs_and_cstr ~schs:[schB';sch3';schC] ~cstr:Constraint.top

let sch5 = parsed "(\\/i=1..n P_i)"
let sch6 = parsed "(/\\i=1..n ~P_i)"
let bck5 = Block.of_schs_and_cstr ~schs:[sch5;sch6] ~cstr:Constraint.top
let sch5' = parsed "(\\/i=1..n-1 P_i)"
let sch6' = parsed "(/\\i=1..n-1 ~P_i)"
let bck5' = Block.of_schs_and_cstr ~schs:[sch5';sch6'] ~cstr:Constraint.top

let test =
  "looping" >:::
    [ "/\\i=1..n P_i" >::
      (fun () ->
        "n-1 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bck1) bck1');
        "n -> n-1"   @? (Lemmas.loops_among (Lemmas.singleton bck1') bck1);
        "n -> n+1"   @? (Lemmas.loops_among (Lemmas.singleton bck2) bck2');
        "n+1 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bck2) bck2);
        "n-1 -> n+1" @? (Lemmas.loops_among (Lemmas.singleton bck2) bck1');
        "n+1 -> n-1" @? (Lemmas.loops_among (Lemmas.singleton bck1') bck2);
        "n-2 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bck1) bck1'');
        "n -> n-2"   @? (Lemmas.loops_among (Lemmas.singleton bck1'') bck1);
      );
    "/\\i=1..n (~P_i \\/ P_i+1)" >::
      (fun () ->
        "n-1 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bck3) bck3');
        "n -> n-1"   @? (Lemmas.loops_among (Lemmas.singleton bck3') bck3);
        "n -> n+1"   @? (Lemmas.loops_among (Lemmas.singleton bck4) bck4');
        "n+1 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bck4) bck4);
        "n-1 -> n+1" @? (Lemmas.loops_among (Lemmas.singleton bck4) bck3');
        "n+1 -> n-1" @? (Lemmas.loops_among (Lemmas.singleton bck3') bck4);
        "n-2 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bck3) bck3'');
        "n -> n-2"   @? (Lemmas.loops_among (Lemmas.singleton bck3'') bck3);
      );
    "[/\\i=1..n ~P_i; P_n+1]" >::
      (fun () ->
        "n-1 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bckA) bckA');
        "n -> n-1"   @? (Lemmas.loops_among (Lemmas.singleton bckA') bckA);
      );
    "[/\\i=1..n (~P_i \\/ P_i+1); ~P_n+1]" >::
      (fun () ->
        "n-1 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bckB) bckB');
        "n -> n-1"   @? (Lemmas.loops_among (Lemmas.singleton bckB') bckB);
      );
    "[P_1;/\\i=1..n (~P_i \\/ P_i+1); ~P_n+1]" >::
      (fun () ->
        "n-1 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bckC) bckC');
        "n -> n-1"   @? (Lemmas.loops_among (Lemmas.singleton bckC') bckC);
      );
    "[(\\/i=1..n P_i);(/\\i=1..n ~P_i)]" >::
      (fun () ->
        "n-1 -> n"   @? (Lemmas.loops_among (Lemmas.singleton bck5) bck5');
        "n -> n-1"   @? (Lemmas.loops_among (Lemmas.singleton bck5') bck5);
        "n -> n"     @? (Lemmas.loops_among (Lemmas.singleton bck5) bck5);
      );
    ]
